@@ -608,7 +608,7 @@ class simulation_model():
 
             # match array pattern and find non-arrayed var
             import re
-            match = re.findall("\[[a-zA-Z1-9,_]*\]", equation)
+            match = re.findall(r'\[[a-zA-Z1-9,_]*\]', equation)
 
             if match:
 
