@@ -28,14 +28,12 @@ def sanitizeName(name):
         .replace(".",'.' )\
         .replace("\"","")\
         .replace("-","")\
+        .replace("'","")\
         .replace(" ","_")
-        # .replace("%","PCT")\
 
     name = re.sub(r"_+","_",name) # Replace multiple underscore with one underscore
     name = re.sub(r"[ ]+", ' ', name) # Many whitespaces to exactly one
 
-    ## Currencies
-    #name = name.replace("$","USD").replace("€","EUR").replace("£","GBP").replace("¥","JPY").lower()
 
     # No leading "."
     if len(name) > 0 and  name[0] == ".": name = name[1:]
