@@ -9,81 +9,119 @@
 # Copyright (c) 2021 transentis labs GmbH
 # MIT License
 
-#from flask import Flask, redirect, url_for, request, make_response, jsonify
-from flask import Flask, request, make_response
-#from BPTK_Py.bptk import bptk  #not needed?
-#import pandas as pd #not needed?
-#import json #TODO not needed?
+from flask import Flask, redirect, url_for, request, make_response, jsonify
+from BPTK_Py.bptk import bptk
+import pandas as pd
+import json
 import copy
 import uuid
-#from json import JSONEncoder #TODO not needed?
+import datetime
+from json import JSONEncoder
+
+
+class InstanceManager:
+    """
+    The class is used to manipulate instances for storing cloned instances, and checking for the session timeout.
+    """
+    def __init__(self, bptk_factory):
+        self._bptk_factory = bptk_factory
+        self._instances = dict()
+
+    def _make_bptk(self):
+        return self._bptk_factory()
+
+    def is_valid_instance(self, instance_uuid):
+        return instance_uuid in self._instances
+
+    def get_instance(self,instance_uuid):
+        if not self.is_valid_instance(instance_uuid):
+            return None
+        # Add the current time to the instances dictionary with its instance id as a key
+        self._update_instance_timestamp(instance_uuid)
+        self._timeout_instances()
+        return self._instances[instance_uuid]["instance"]
+
+    def _update_instance_timestamp(self, instance_uuid):
+        if self.is_valid_instance(instance_uuid):
+            self._instances[instance_uuid]["time"]= datetime.datetime.now()
+
+    def create_instance(self):
+        """
+        The method generates a universally unique identifier in hexadecimal, that is used as key for the instances.
+
+        Returns: String
+            The uuid value generated for the current instance.
+        """
+
+        self._timeout_instances()
+
+        instance_uuid = uuid.uuid1().hex
+        self._instances[instance_uuid] = dict()
+        self._instances[instance_uuid]["instance"] = self._make_bptk()
+        self._instances[instance_uuid]["time"] = dict()
+
+        return instance_uuid
+
+    def _timeout_instances(self):
+        """
+        The method checks for the session timeout, and deletes the instance if it is.
+
+        Returns:
+            True: Boolean.
+                Means that the specified time has already passed, and the session should be terminated.
+        """
+
+        timeout = datetime.timedelta(minutes=5)  # Terminate the session after 5 minutes
+
+        for key in self._instances.keys():
+            current_time = datetime.datetime.now()
+            last_call_time = self._instances[key]["time"]
+            if last_call_time:
+                if current_time >= last_call_time + timeout:
+                    del self._instances[key]
+
+######################
+##  REST API CLASS  ##
+######################
+
 
 class BptkServer(Flask):
     """
-    This class provides a Flask-based server that provides a REST-API for running bptk scenarios. Currently only SD scenarios are supported (both SD DSL and XMILE).
-    
-    The class inherts the properties and methods of Flask and doesn't expose any further public methods.
-
-    The default port of the server is 5000.
-    
-    Initialize the server with the import name and the bptk.
-
-    Parameters:
-        import_name: String.
-            The name of the application package. Usually __name__. This helps locate the root_path for the blueprint.
-        bptk: bptk Object.
-            An instance of bptk that contains all relevant scenarios. These are then made available via REST endpoints.
+    This class provides a Flask-based server that provides a REST-API for running bptk scenarios. The class inherts the properties and methods of Flask and doesn't expose any further public methods.
     """
-    
-    def __init__(self, import_name, bptk):
+
+    def __init__(self, import_name, bptk_factory=None):
+        """
+        Initialize the server with the import name and the bptk.
+        :param import_name: the name of the application package. Usually __name__. This helps locate the root_path for the blueprint.
+        :param bptk: simulations made by the bptk.
+        """
         super(BptkServer, self).__init__(import_name)
-        self.bptk = bptk
-        self.bptk_copy = copy.deepcopy(self.bptk)
-        self.instances_dict = dict()
+        self._bptk = bptk_factory() if bptk_factory is not None else None
+        
+        
+        self._instance_manager = InstanceManager(bptk_factory)
         # specifying the routes and methods of the api
         self.route("/", methods=['GET'])(self._home_resource)
-        self.route("/run", methods=['PUT','POST'], strict_slashes=False)(self._run_resource)
+        self.route("/run", methods=['POST', 'PUT'], strict_slashes=False)(self._run_resource)
         self.route("/scenarios", methods=['GET'], strict_slashes=False)(self._scenarios_resource)
-        self.route("/equations", methods=['PUT','POST'], strict_slashes=False)(self._equations_resource)
-        self.route("/agents", methods=['PUT','POST'], strict_slashes=False)(self._agents_resource)
-        #self.route("/start-instance", methods=['POST'], strict_slashes=False)(self._start_instance_resource)
-        #self.route("/<instance_id>/run-step", methods=['POST', 'PUT'], strict_slashes=False)(self._run_step_resource)
-        
+        self.route("/equations", methods=['POST'], strict_slashes=False)(self._equations_resource)
+        self.route("/agents", methods=['POST', 'PUT'], strict_slashes=False)(self._agents_resource)
+        self.route("/start-instance", methods=['POST'], strict_slashes=False)(self._start_instance_resource)
+        self.route("/<instance_uuid>/run-step", methods=['POST'], strict_slashes=False)(self._run_step_resource)
+        self.route("/<instance_uuid>/begin-session", methods=['POST'], strict_slashes=False)(self._begin_session_resource)
+        self.route("/<instance_uuid>/end-session", methods=['POST'], strict_slashes=False)(self._end_session_resource)
+
     def _home_resource(self):
         """
         The root endpoint returns a simple html page for test purposes.
-
-        Returns:
-             Returns the HTML string "<h1>BPTK-Py REST API Server</h1>".
-
         """
         return "<h1>BPTK-Py REST API Server</h1>"
 
     def _run_resource(self):
-        """Runs a s set of scenarios with given settings and returns the results.
-
-        You can override the scenario settings, thus allowing interactive simulations.
-
-        This endpoint runs the scenario through all timesteps.
-
-        Parameters:
-            scenario managers: List.
-                JSON list of scenario managers that should be run.
-            scenarios: List.
-                JSON list of scenarios that should be run.
-            equations: List.
-                JSON list of equations that data should be retrieved for.
-            settings: Dictionary.
-                JSON dictionary that has scenario managers as the key. For each scenario manager there is a dictionary of scenarios, which contains a dictionary of constants and their settings.
-        
-        Returns:
-            A JSON dictionary that contains the simulation results ordered by scenario_manager, scenario, equations. The equations dictionary contains the timesteps as keys with the simulation results as values.
         """
- 
-        
-        self.logger.info("Request is JSON: {}".format(request.is_json))
-        self.logger.info("Request is JSON: {}".format(request.data))
-
+        Given a JSON dictionary that defines the relevant simulation scenarios and equations, this endpoint runs those scenarios and returns the data generated by the simulations.
+        """
         if not request.is_json:
             resp = make_response('{"error": "please pass the request with content-type application/json"}',500)
             resp.headers['Content-Type'] = 'application/json'
@@ -94,10 +132,10 @@ class BptkServer(Flask):
 
         try:
             settings = content["settings"]
-        
+
             for scenario_manager_name, scenario_manager_data in settings.items():
                 for scenario_name, scenario_settings in scenario_manager_data.items():
-                    scenario = self.bptk.get_scenario(scenario_manager_name,scenario_name)
+                    scenario = self._bptk.get_scenario(scenario_manager_name,scenario_name)
                     if "constants" in scenario_settings:
                         constants = scenario_settings["constants"]
                         for constant_name, constant_settings in constants.items():
@@ -106,14 +144,12 @@ class BptkServer(Flask):
                         points = scenario_settings["points"]
                         for points_name, points_settings in points.items():
                             scenario.points[points_name]=points_settings
-                    self.bptk.reset_scenario_cache(scenario_manager=scenario_manager_name,scenario=scenario_name)
+                    self._bptk.reset_scenario_cache(scenario_manager=scenario_manager_name,scenario=scenario_name)
         except KeyError:
-            self.logger.info("Settings not specified")
             pass
 
-
         try:
-            scenario_managers=content["scenario_managers"]
+            scenario_managers = content["scenario_managers"]
         except KeyError:
             resp = make_response('{"error": "expecting scenario_managers to be set"}',500)
             resp.headers['Content-Type']='application/json'
@@ -121,28 +157,27 @@ class BptkServer(Flask):
             return resp
 
         try:
-            scenarios=content["scenarios"]
+            scenarios = content["scenarios"]
         except KeyError:
-            resp = make_response('{"error": "expecting scenario_managers to be set"}',500)
+            resp = make_response('{"error": "expecting scenario_managers to be set"}', 500)
             resp.headers['Content-Type']='application/json'
             resp.headers['Access-Control-Allow-Origin']='*'
             return resp
 
         try:
-            equations=content["equations"]
+            equations = content["equations"]
         except KeyError:
-            resp = make_response('{"error": "expecting equations to be set"}',500)
+            resp = make_response('{"error": "expecting equations to be set"}', 500)
             resp.headers['Content-Type']='application/json'
             resp.headers['Access-Control-Allow-Origin']='*'
             return resp
 
-
-        result = self.bptk.run_scenarios(
-              scenario_managers=scenario_managers,
-              scenarios=scenarios,
-              equations=equations,
-              return_format="json"
-            )
+        result = self._bptk.run_scenarios(
+            scenario_managers=scenario_managers,
+            scenarios=scenarios,
+            equations=equations,
+            return_format="json"
+        )
 
         if result is not None:
             resp = make_response(result, 200)
@@ -152,54 +187,48 @@ class BptkServer(Flask):
         resp.headers['Content-Type'] = 'application/json'
         resp.headers['Access-Control-Allow-Origin']='*'
         return resp
-    
+
     def _scenarios_resource(self):
-        """Returns all available scenarios for the current simulation.
-
-        Returns:
-            A JSON dictionary containing the scenario manager names as keys and a list of scenario names as the values. 
-
+        """
+        The endpoint returns all available scenarios for the current simulation.
         """
 
-        scenarios = self.bptk.get_scenario_names(format="dict")
+        scenarios = self._bptk.get_scenario_names(format="dict")
 
         if not scenarios:
             resp = make_response('{"error": "expecting the model to have scenarios"}',500)
             resp.headers['Content-Type'] = 'application/json'
             resp.headers['Access-Control-Allow-Origin']='*'
             return resp
-        
+
         resp = make_response(scenarios, 200)
 
         return resp
-    
+
     def _equations_resource(self):
-        """Returns all available equations given the name of a scenario manager and of a scenario.
-
-        Parameters:
-            A JSON dictionary with the keys "scenarioManager" and "scenario".
-
-        Returns:
-            A JSON dictionary with the keys "constants", "converters","flows", "points" and "stocks". Each key contains a list of the scenario's equations.
-
         """
-        
+        This endpoint returns all available equations given the name of a scenario manager and of a scenario.
+        """
+
         if not request.is_json:
             resp = make_response('{"error": "please pass the request with content-type application/json"}',500)
             resp.headers['Content-Type'] = 'application/json'
             resp.headers['Access-Control-Allow-Origin']='*'
             return resp
-    
+
         content = request.get_json()
-        
-        try:
-            scenario_manager_name = content["scenarioManager"]
-        except KeyError:
-            resp = make_response('{"error": "expecting scenarioManager to be set"}',500)
+
+        if ("scenarioManager" not in content) and ("scenario_manager" not in content):
+            resp = make_response('{"error": "expecting scenarioManager or scenario_manager to be set"}',500)
             resp.headers['Content-Type']='application/json'
             resp.headers['Access-Control-Allow-Origin']='*'
             return resp
-        
+
+        if "scenario_manager" in content:
+            scenario_manager_name=content["scenario_manager"]
+        else:
+            scenario_manager_name=content["scenarioManager"]
+
         try:
             scenario_name = content["scenario"]
         except KeyError:
@@ -207,16 +236,15 @@ class BptkServer(Flask):
             resp.headers['Content-Type']='application/json'
             resp.headers['Access-Control-Allow-Origin']='*'
             return resp
-            
-  
-        scenario = self.bptk.get_scenario(scenario_manager_name,scenario_name)
+
+        scenario = self._bptk.get_scenario(scenario_manager_name,scenario_name)
 
         equations_names = {}
         stocks_names = set()
         flows_names = set()
         converters_names = set()
         constants_names = set()
-        points_names = set()        
+        points_names = set()
 
         for equation in sorted(scenario.model.stocks):
             stocks_names.add(equation)
@@ -241,15 +269,18 @@ class BptkServer(Flask):
             resp = make_response('{"error": "no data was returned from simulation"}', 500)
 
         return resp
-    
+
     def _agents_resource(self):
-        
+        """
+        For an agent-based or hybrid model, this endpoint returns all the agents in the model with their corresponding states and properties.
+        """
+
         if not request.is_json:
             resp = make_response('{"error": "please pass the request with content-type application/json"}',500)
             resp.headers['Content-Type'] = 'application/json'
-            resp.headers['Access-Control-Allow-Origin']='*'
+            resp.headers['Access-Control-Allow-Origin'] = '*'
             return resp
-        
+
         content = request.get_json()
         try:
             scenario_manager_name = content["scenarioManager"]
@@ -258,7 +289,7 @@ class BptkServer(Flask):
             resp.headers['Content-Type']='application/json'
             resp.headers['Access-Control-Allow-Origin']='*'
             return resp
-        
+
         try:
             scenario_name = content["scenario"]
         except KeyError:
@@ -267,14 +298,14 @@ class BptkServer(Flask):
             resp.headers['Access-Control-Allow-Origin']='*'
             return resp
 
-        scenario = self.bptk.get_scenario(scenario_manager_name,scenario_name)
-        
+        scenario = self._bptk.get_scenario(scenario_manager_name,scenario_name)
+
         if not scenario.model.agents: # Checking if the model has agents
             resp = make_response('{"error": "expecting the model to have agents"}',500)
             resp.headers['Content-Type'] = 'application/json'
             resp.headers['Access-Control-Allow-Origin']='*'
             return resp
-        
+
         agents_dict = dict()
 
         for agent in scenario.model.agents:
@@ -284,65 +315,51 @@ class BptkServer(Flask):
             agents_dict[agent_name]["states"] = [agent_state]
             agent_properties = list(agent.properties.keys())
             agents_dict[agent_name]["properties"] = agent_properties
-            
+
         if agents_dict is not None:
             resp = make_response(agents_dict, 200)
         else:
             resp = make_response('{"error": "no data was returned from simulation"}', 500)
 
         return resp
- 
+
     def _start_instance_resource(self):
         """
         This endpoint starts a new instance of BPTK on the server side, so that simulations can run in a "private" session. The endpoint returns an instance_id, which is needed to identify the instance in later calls.
         """
-        
-        # Create a universally unique identifier in hex to store 
-        cloned_bptk_uuid = uuid.uuid1().hex
-        
-        self.instances_dict[cloned_bptk_uuid] = self.bptk_copy
-        
-        if cloned_bptk_uuid is not None:
-            resp = make_response(cloned_bptk_uuid, 200)
+
+        # store the new instance in the instance dictionary.
+        instance_uuid = self._instance_manager.create_instance()
+
+        if instance_uuid is not None:
+            resp = make_response(f'{{"instance_uuid":"{instance_uuid}"}}', 200)
         else:
-            resp = make_response('{"error": "no data was returned from simulation"}', 500)
+            resp = make_response('{"error": "instance could not be started"}', 500)
 
         return resp
-    
-    def _run_step_resource(self, instance_id):
+
+    def _begin_session_resource(self, instance_uuid):
+        """This endpoint starts a session for single step simulation. There can only be one session per instance at a time.
+        Currently only System Dynamics scenarios are supported for both SD DSL and XMILE models.
         """
-        This endpoint advances the relevant scenarios by one timestep and returns the data for that timestep.
-        """
-        
+
         if not request.is_json:
-            resp = make_response('{"error": "please pass the request with content-type application/json"}',500)
+            resp = make_response('{"error": "please pass the request with content-type application/json"}', 500)
             resp.headers['Content-Type'] = 'application/json'
-            resp.headers['Access-Control-Allow-Origin']='*'
+            resp.headers['Access-Control-Allow-Origin'] = '*'
+            return resp
+
+        # Checking if the instance id is valid.
+        if not self._instance_manager.is_valid_instance(instance_uuid):
+            resp = make_response('{"error": "expecting a valid instance id to be given"}', 500)
+            resp.headers['Content-Type'] = 'application/json'
+            resp.headers['Access-Control-Allow-Origin'] = '*'
             return resp
 
         content = request.get_json()
 
         try:
-            settings = content["settings"]
-
-            for scenario_manager_name, scenario_manager_data in settings.items():
-                for scenario_name, scenario_settings in scenario_manager_data.items():
-                    scenario = self.bptk.get_scenario(scenario_manager_name,scenario_name)
-                    if "constants" in scenario_settings:
-                        constants = scenario_settings["constants"]
-                        for constant_name, constant_settings in constants.items():
-                            scenario.constants[constant_name]=constant_settings
-                    if "points" in scenario_settings:
-                        points = scenario_settings["points"]
-                        for points_name, points_settings in points.items():
-                            scenario.points[points_name]=points_settings
-                    self.bptk.reset_scenario_cache(scenario_manager=scenario_manager_name,scenario=scenario_name)
-        except KeyError:
-            self.logger.info("Settings not specified")
-            pass
-        
-        try:
-            scenario_managers=content["scenario_managers"]
+            scenario_managers = content["scenario_managers"]
         except KeyError:
             resp = make_response('{"error": "expecting scenario_managers to be set"}',500)
             resp.headers['Content-Type']='application/json'
@@ -350,35 +367,86 @@ class BptkServer(Flask):
             return resp
 
         try:
-            scenarios=content["scenarios"]
+            scenarios = content["scenarios"]
         except KeyError:
-            resp = make_response('{"error": "expecting scenario_managers to be set"}',500)
+            resp = make_response('{"error": "expecting scenarios to be set"}', 500)
             resp.headers['Content-Type']='application/json'
             resp.headers['Access-Control-Allow-Origin']='*'
             return resp
 
         try:
-            equations=content["equations"]
+            equations = content["equations"]
         except KeyError:
-            resp = make_response('{"error": "expecting equations to be set"}',500)
+            resp = make_response('{"error": "expecting equations to be set"}', 500)
             resp.headers['Content-Type']='application/json'
             resp.headers['Access-Control-Allow-Origin']='*'
             return resp
-        
-        # getting the results for the first time step only
-        result = self.bptk.plot_scenarios(
-              scenario_managers=scenario_managers,
-              scenarios=scenarios,
-              equations=equations,
-              visualize_to_period = 1,
-              return_df=True
-            )
+
+        instance = self._instance_manager.get_instance(instance_uuid)
+
+        instance.begin_session(
+            scenario_managers=scenario_managers,
+            scenarios=scenarios,
+            equations=equations
+        )
+
+        resp = make_response('{"msg":"session started"}', 200)
+        resp.headers['Content-Type'] = 'application/json'
+        resp.headers['Access-Control-Allow-Origin']='*'
+        return resp
+
+    def _end_session_resource(self, instance_uuid):
+        """This endpoint ends a session for single step simulation and resets the internal cache.
+        """
+        # Checking if the instance id is valid.
+        if not self._instance_manager.is_valid_instance(instance_uuid):
+            resp = make_response('{"error": "expecting a valid instance id to be given"}', 500)
+            resp.headers['Content-Type'] = 'application/json'
+            resp.headers['Access-Control-Allow-Origin'] = '*'
+            return resp
+
+        instance = self._instance_manager.get_instance(instance_uuid)
+        instance.end_session()
+
+        resp = make_response('{"msg":"session terminated"}', 200)
+        resp.headers['Content-Type'] = 'application/json'
+        resp.headers['Access-Control-Allow-Origin']='*'
+        return resp
+
+    def _run_step_resource(self, instance_uuid):
+        """
+        This endpoint advances the relevant scenarios by one timestep and returns the data for that timestep.
+
+        Arguments:
+            settings: JSON
+                Dictionary structure with a key "settings" that contains the settings to apply for that step. These can be constants and points.
+        """
+        # Checking if the instance id is valid.
+        if not self._instance_manager.is_valid_instance(instance_uuid):
+            resp = make_response('{"error": "expecting a valid instance id to be given"}', 500)
+            resp.headers['Content-Type'] = 'application/json'
+            resp.headers['Access-Control-Allow-Origin'] = '*'
+            return resp
+
+        instance = self._instance_manager.get_instance(instance_uuid)
+
+        if not request.is_json:
+            result = instance.run_step()
+        else:
+            content = request.get_json()
+            if "settings" in content:
+                result = instance.run_step(settings=content["settings"])
+            else:
+                resp = make_response('{"error": "expecting settings to be set"}', 500)
+                resp.headers['Content-Type'] = 'application/json'
+                resp.headers['Access-Control-Allow-Origin'] = '*'
+                return resp
 
         if result is not None:
-            resp = make_response(result.to_json(), 200)
+            resp = make_response(json.dumps(result), 200)
         else:
-            resp = make_response('{"error": "no data was returned from simulation"}', 500)
+            resp = make_response('{"error": "no data was returned from run_step"}', 500)
 
         resp.headers['Content-Type'] = 'application/json'
         resp.headers['Access-Control-Allow-Origin']='*'
-        return resp
\ No newline at end of file
+        return resp
